#!/usr/bin/env python
## -*- coding: utf-8 -*-
#
#Combine two files, where each file consists of two tab-delimited columns,
# and the second column of the left-hand file contains the same tokens
# as the first column of the right-hand file.
#
# The second column is assumed to be the underlying representation of a word,
# hence the name of this file: MergeByUR.
#
# Both files are assumed to be sorted by the common column.  However,
# it is NOT assumed that the two files have the same number of lines.
# In particular, the left-hand file may have multiple lines with different
# first tokens but the same second token (i.e. multiple surface form from
# a single underlying form, due to optional phonological rules), or
# multiple lines with the same first token but different second tokens
# (ambiguity in parsing).  Similar things may happen in the right-hand
# file.
#
# However, it should be the case that both files contain the same number of
# *unique* tokens in the shared column.

# Arguments: -l <LeftFile>
#            -r <RightFile>
#            -u (=> UTF-8)
# The first two arguments are obligatory, the third is optional (and if
# not given, the input is assumed to be non-UTF8, presumably ISO).

# Input: The two files, as described above.
# Output: A sequence of lines output to stdout, joined by the common token.
#         The resulting three columns are tab delimited.
#         For any tokens which appear more than once in the second input file,
#         the corresponding token of the first input file will be repeated.
# For example, if the two files are:
#    | abc  ABC || ABC alpha |
#    | def  DEF || ABC beta  |
#    |          || DEF gamma |
# then the output will be:
#    | abc ABC alpha |
#    | abc ABC beta  |
#    | def DEF gamma |
# There should never be a case where the right-hand token of the first file
# does not have a correspondent in the (next line of) the second file.  If
# we encounter such a case, we will output a line in which the missing third
# column is '**error**'.  There should also never be a case where the left-hand
# token of the second file does not have a correspondent in the first file.
# If such a case arises, we output a line in which the missing first column
# is '**error**'.  In either case, we exit with an exit code of -1.

import sys
import getopt
import codecs   #For UTF-8


def OpenInput():
    "Open the two input files (output is to stdout)."

    #Default is non-Unicode:
    bUTF8 = False

    #Look for command line args:
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'l:r:u:')
    except getopt.GetoptError:
        print 'Usage: MergeByUR.py -l <Left-hand file> -o <Right-hand file> (-u)'
        sys.exit(1)

    for o in opts:
        if o[0] == '-l':
            LFile = o[1]
        elif o[0] == '-r':
            RFile = o[1]
        elif o[0] == '-u':
            bUTF8 = True
            (UTF8Encode, UTF8Decode, UTF8Reader, UTF8Writer) = codecs.lookup('utf-8')
               #In case we need to read/write UTF-8.
               # We don't use the UTF8Encode or UTF8Decode, but they are returned by codecs.lookup()

    #Make sure both input files exist:
    try:
        if bUTF8:
            LStream = UTF8Reader(open(LFile, 'r'))
        else:
            LStream = open(LFile, 'r')
    except IOError:
        print "Error: Cannot open file %s" %LFile
        sys.exit(1)

    try:
        if bUTF8:
            RStream = UTF8Writer(open(RFile, 'r'))
        else:
            RStream = open(RFile, 'r')
    except IOError:
        print "Error: Cannot open file %s" %RFile
        sys.exit(1)

    return (LStream, RStream)

def ParseLine(sLine):
    """A trivial function to parse sLine into two parts at a tab char,
       removing any other initial or final whitespace.
       Return the two parts, or -1 if the line could not be split."""
    try:
        (sPart1, sPart2) = map(lambda(s): s.strip(), sLine.split("\t", 1))
        return (sPart1, sPart2)
    except:
        return -1


#--------------MAIN--------------
iErrorStatus = 0  #Hope for the best

#--------------Process records:
#Init the pipeline:
(LStream, RStream) = OpenInput()
LLine = LStream.readline()
(LToken1, LToken2) = ParseLine(LLine)
RLine = RStream.readline()
(RToken1, RToken2) = ParseLine(RLine)

#Now start matching lines, until we run out of lines in the first file:
while LLine <> "" and RLine <> "":
    #First collect all the token pairs from the left-hand file whose
    # second token is the same (usually this will be just a single line,
    # but where there is ambiguity it will be more than one).  Start
    # with the first record:
    LKeyToken = LToken2
    LTokenList = [LToken1]
    #Get any additional lines from the left-hand file whose second token matches:
    LLine = LStream.readline()
    if LLine != "":
        (LToken1, LToken2) = ParseLine(LLine)
        while LKeyToken == LToken2:
            LTokenList = LTokenList + [LToken1]
            LLine = LStream.readline()
            if LLine == "":
                break
            else:
                (LToken1, LToken2) = ParseLine(LLine)
            #When we exit this while-loop, (LToken1, LToken2) will be
            # the first token pair from the left-hand file
        # for the next major loop.

    #Next collect all the token pairs from the right-hand file whose
    # middle token matches the KeyToken.  Again, this is likely to be
    # just one, but it may be more.
    #First line:
    RKeyToken = RToken1
    RTokenList = [RToken2]
    #...and any additional lines from the right-hand file whose first token matches:
    RLine = RStream.readline()
    if RLine != "":
        (RToken1, RToken2) = ParseLine(RLine)
        while RKeyToken == RToken1:
            RTokenList = RTokenList + [RToken2]
            RLine = RStream.readline()
            if RLine == "":
                break
            else:
                (RToken1, RToken2) = ParseLine(RLine)
            #When we exit this while-loop, (RLToken1, RToken2) will be
            # the first token pair from the right-hand file for
            # the next major loop.

    if LKeyToken != RKeyToken:
        #Error condition
        sys.stderr.write("***Error: %s =\= %s\n" %(LKeyToken, RKeyToken))
        iErrorStatus = -1
    else:
        #For each left-most token, give every parse into the middle token:
        for LeftMostToken in LTokenList:
            #...and each parse into the right-most token:
            for RightMostToken in RTokenList:
                sys.stdout.write("%s\t%s\t%s\n" %(LeftMostToken, LKeyToken, RightMostToken))

#The left-hand file is now all used up.  We could verify that the right-hand file is used
# up too...
sys.exit(iErrorStatus)
