#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#Convert the stems as stored in an XML lexicon file
# into separate files for each POS.
# There are some special treatments of POSs, subclass strings
# (the lexicon uses multiple ones, we reduce this to a smaller set),
# and the form (when this contains a colon, used to mark vowel length).
# (A UID field may be added later, in place of the gloss field.)
#
# Arguments: -c <CodesFile> (Obligatory)
#            -i <InputFile> (optional, defaults to stdin)
#            -o <OutputFile> (optional, defaults to stdout)
#            -e <ErrFile> (optional, and not currently used; defaults to stderr)
# Input: An XML file; the fields we're interested in are listed in <CodesFile>.
#        Fields:
#            (1) Target language word: tags given in sLexemeTags
#            (2) POS: tags given in sPOSTags
#            (3) Subclass (may be empty): tags given in sSubclassTags
#            (4) Allomorph code (may be empty or repeat): tags given in sAllomorphTags
# Output: We create separate files in the current dir in lexc format for each
#        POS found in the input that corresponds to an acceptable POS (the list
#        of acceptable POSs is given <CodesFile>.  We make some modifications,
#        removing parenthesized material at the end--this usually corresponds
#        to subcats, probably irrelevant to the morphology.  The separate files
#        are so that each POS can be loaded into separate xfst variables.
#        Each such file is named <POS>.lexc  (There is a special case for
#        Aux, because Windows/DOS makes it nearly impossible to work with
#        a file whose basename is Aux.)
#        In addition, we create single xfst file named as specified in the
#        -o arg, with the xfst cmds to load the lexc files and 'defines'
#        for each lexicon.  (But see 'To-do' below.)
#        For verbs, the subclass (paradigm class) is turned into a string preceded
#        by '^' and attached (without a space) to the upper (gloss) and
#        lower (language form) field.
#        (No attempt is made to sort the entries within a given Verbal subcategory by class.)
#        For nouns, the subclass is concatenated to the POS ("N"), and the two
#        are used jointly for morphotactics.



import sys
import getopt
import time
import string
import codecs   #For UTF-8
import xml.dom.minidom
from tempfile import mkdtemp


#----------------------Functions--------------------

def OpenInputOutput():
    #Look for command line args:
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'c:i:o:e:')
    except getopt.GetoptError:
        sys.stderr.write('Usage: NahTab2MultLexc.py -c <CodesFile> (-i <InputFile>) (-o <OutputFile>) (-e <ErrorFile>)\n')
        sys.exit(1)

    #Defaults:
    InFile  = 'stdin'
    OutFile = 'stdout'
    ErrFile = 'stderr'

    #Non-defaults:
    CodesFile = ''  #No default for this
    for o in opts:
        if o[0]   == '-c':
            CodesFile = o[1]
        elif o[0] == '-i':
            InFile  = o[1]
        elif o[0] == '-o':
            OutFile = o[1]
        elif o[0] == '-e':
            ErrFile = o[1]

    #Make sure we got a CodesFile:
    if CodesFile == '':
        sys.stderr.write('Usage: NahTab2MultLexc.py -c <CodesFile> (-i <InputFile>) (-o <OutputFile>) (-e <ErrorFile>)\n')
        sys.exit(1)

    (UTF8Encode, UTF8Decode, UTF8Reader, UTF8Writer) = codecs.lookup('utf-8')
       #We don't use the UTF8Encode or UTF8Decode, but they are returned by codecs.lookup()

    #Now open the various files:
    #Make sure the input stream exists, and open it (note that we do NOT need to open
    # it as a Unicode stream, because the minidom parser takes care of that):
    try:
        if InFile == 'stdin':
            InStream = sys.stdin
        else:
            InStream = open(InFile, 'r')
    except IOError:
        sys.stderr.write("Error: Cannot open file %s" %InFile)
        sys.exit(1)

    if OutFile == 'stdout':
        OutStream = UTF8Writer(sys.stdout)
    else:
        OutStream = UTF8Writer(open(OutFile, 'w'))

    if ErrFile == 'stderr':
        ErrStream = UTF8Writer(sys.stderr)
    else:
        ErrStream = UTF8Writer(open(ErrFile, 'w'))

    return (CodesFile, InStream, OutStream, ErrStream)


def isLegalChars(sString):
    "Return True if sString consists entirely of chars in the designated alphabet, \
     False otherwise.  Used to test stems for validity."
    for cChar in sString:
        if not cChar in sAlphabet:
            return False
    #If we get here, everything's hunky-dory:
    return True


#--------------------------Class Lexeme---------------------------
class Lexeme:
    """Attributes:
          sCitationForm        (String)
          sUnderlyingForm      (String): usually identical to sCitationForm, but not always
          POS                  (Object, defined in this file)
          sParadigmClass       (String): for nouns, e.g."@P.STRATUM.SPANISH@"; default is an empty string
          sStemClass           (String): for stem classes in verbs, e.g. "^C2b"; default is an empty string
          sAllomorphClasses    (String): for other allomorphy in nouns and verbs, e.g. "^h=l" or "^iEpen";
                                         default is an empty string
       Possibly to be added later:
          UID
       All of the work is in the __init__() method, which parses an input string into
       the various fields, performs error checking, and does normalization on the fields.
       Under certain circumstances (like a bad POS), the init will return None;
       otherwise, it returns self."""

    def __init__(self, domLexemeNode):
        """Create a normalized version of domLexemeNode if it is valid.  Some things we test
           for are the chars in the stem, spaces in the POS or conjugation class
           (which are often comments), backslashes in the line, etc.  If any of the fields
           in domLexemeNode are invalid, we raise an exception for the first error we find.
           (But exceptions are treated as warnings, i.e. we continue, usually with the next record.)
        """

        #Initializations (we don't have a self.domLexemeNode, because we want to be able to
        # garbage collect those before we're done with this object):
        self.sCitationForm     = None
        self.sUnderlyingForm   = None
        self.POS               = None
        self.sStemClass        = None   #Temp value; a value of '' will be used for no stem class
        self.sParadigmClass    = None   #Temp value; a value of '' will be used for no paradigm class
        self.sAllomorphClasses = None   #Temp value; a value of '' will be used for no allomorph classes
        self.FlagDiacritics    = ''

        #Create an error msg string containing this Lexeme, should we need it.
        #Init to a space:
        self.sErrorMsgString = ' '
        #...then add the relevant tags (CitationForms, POS, Subclass, and Allomorphs):
        for sTag in sAllTags:
            #Create a string containing all the text fields marked by this tag:
            sField = ''
            for domFieldNode in domLexemeNode.getElementsByTagName(sTag):
                #Loop over fields marked by this tag (usually only one, but multiple ones for e.g. allomorphs):
                sText = ''
                for domTextNode in filter(lambda(N): N.nodeType == xml.dom.minidom.Node.TEXT_NODE, domFieldNode.childNodes):
                    sText = sText + ' ' + domTextNode.nodeValue
                sField = sField + '  ' + sText
            self.sErrorMsgString = self.sErrorMsgString + '\t' + sField
        self.sErrorMsgString = self.sErrorMsgString + '\n'

        #Citation form of stem:
        sRawStem = self.GetFirstMatchingTag(domLexemeNode, sCitationFormTags, Lexeme.MissingStemForm)
        if sRawStem == None:
            return None  #GetFirstMatchingTag() already raised a warning
        #If we get here, there was a stem; see if it's any good:
        if sRawStem == '':
            raise Lexeme.EmptyStemWarning, self.sErrorMsgString
            return None
        #Remove any leading or trailing chars that it's legal to remove (such as leading or trailing
        # whitespace):
        sRawStem = sRawStem.strip(sCharactersToStrip)
        #Then check remainder to ensure that it isn't a phrasal lexeme (one containing a space char):
        if sRawStem.find(' ') != -1:
            raise Lexeme.PhrasalLexemeWarning, self.sErrorMsgString
            return None
        #...and that it contains only chars that are in the alphabet (we could do both checks
        # simultaneously, but it's nice to have a separate readout of the two conditions: the latter,
        # unlike the former, is indicative of an error):
        if not isLegalChars(sRawStem):
            #Some entries have '-----', we don't want those; this also catches phrasal stems,
            # which would have a space char, as well as some bad spellings.
            raise Lexeme.StemCharsWarning, self.sErrorMsgString
            return None
        else:
            #Stem is OK, use it:
            self.sCitationForm = self.NormalizeStem(sRawStem)

        #POS:
        sRawPOS = self.GetFirstMatchingTag(domLexemeNode, sPOSTags, Lexeme.MissingPOSWarning)
        self.POS = self.NormalizePOS(sRawPOS)  #Return value will either be a POS object or None
        if self.POS == None:
            #If we get here, the POS was not in the list of legit POSs (dictPOSNames), or there was
            # no POS; we don't want to keep this lex entry, so quit:
            return None

        #Stem class (verbs only):
        sRawStemClass = self.GetFirstMatchingTag(domLexemeNode, sStemClassTags)
            #Stem class is optional, hence no warning for not having one

        #Underlying form:
        # set this using the _raw_ stem class (if any):
        self.SetUnderlyingForm(sRawStemClass)
            #This always succeeds (no error msgs).

        #Stem Class:
        self.sStemClass = self.NormalizeStemClass(sRawStemClass)

        #Allomorph class:
        # These exist for some nouns and verbs.
        sRawAllomorphClass = self.GetFirstMatchingTag(domLexemeNode, sAllomorphTags)
            #Allomorph class is optional, hence no warning for not having one.
            # Some lexemes have > 1 allomorphy code.
        self.sAllomorphClasses = self.NormalizeAllomorphClass(sRawAllomorphClass)
            #An error in this is ignored (since other allomorphs will be OK)

        #Now set the lexeme's paradigm class (this is really only relevant to nouns):
        sRawParadigmClass = self.GetFirstMatchingTag(domLexemeNode, sParadigmTags)
        self.sParadigmClass = self.NormalizeParadigmClass(sRawParadigmClass)

        #Subclass (relevant only for nouns; verb subclasses are already given in
        # their POS field).
        sRawSubclass = self.GetFirstMatchingTag(domLexemeNode, sSubclassTags)
            #Subclass is optional, hence no warning for not having one
        self.POS = self.ModPOSwithSubclass(sRawSubclass)

        #Irregular forms (this only handles a few of the cases now):
        sIrregularities = self.GetFirstMatchingTag(domLexemeNode, sIrregularTags)
#        if sIrregularities != '':
#            ErrOut.write("self.sCitationForm = %s, sIrregularities = '%s'\n" %(self.sCitationForm, sIrregularities))  ##DEBUG
            #There may eventually be more than one of these, in which case GetFirstMatchingTag()
            # won't work.
        #For now, we're only dealing with nouns that lack a plural form; this is represented
        # by 'Singular' in the \plural field.  (See also my email of 29 Apr 07 9:44 PM.)
        self.InterpretIrregularities(sIrregularities)

        #If we get here, everything checks out; add it to the list of lexemes for its POS
        # (for uninflected forms, this will be a dummy POS):
        self.POS.Lexemes.append(self)


    def GetFirstMatchingTag(self, domLexemeNode, sTags, ErrorMsg = None):
        """Get the text from the element matching the first of sTags that we find.
           If we don't find any such element, then raise ErrorMsg if there is one,
           and return None.  (There won't be an error msg if the element is optional.)"""
        for sTag in sTags:
            #See if the lexeme uses this tag:
            domFieldNodes = domLexemeNode.getElementsByTagName(sTag)
            if domFieldNodes != []:
                #If we get here, the lexeme does contain at least one instance of that tag.
                # Cat together all the text from all instances of that tag into sFields:
                sFields = ''
                for domFieldNode in domFieldNodes:
                    #...by first catting together all the text from each individual text node
                    # (there will probably be only one such text node, but just in case):
                    sField = ''
                    for domTextNode in filter(lambda(N): N.nodeType == xml.dom.minidom.Node.TEXT_NODE, domFieldNode.childNodes):
                        sField = sField + ' ' + domTextNode.nodeValue
                    #...and then add the individual text node's text to the accumulator:
                    sFields = sFields + '  ' + sField
                #And since we want to return the node matching the first tag we find,
                # break out of the outer for-loop and return this value:
                return sFields.strip()
        #If we get here, we ran out of tags without finding a match, meaning there wasn't
        # any matching element in this lex entry.
        #If we were given an error msg, then raise it (there will not be an error msg if
        # the field is optional):
        if ErrorMsg != None:
            raise ErrorMsg, self.sErrorMsgString
        return None



    #-------------------Normalization functions-----------------------
    #These are called by the __init__() function.  They all return a normalized string
    # or object, or None if the input was null.  Setting the corresponding variable
    # in self is the duty of the calling code.

    def NormalizeStem(self, sRawStem):
        """Normalize the citation form stem.  Specifically, colons
           (used to mark vowel length) get escaped with a '%' character, so lexc
           doesn't complain."""
        return sRawStem.replace(":", "%:")


    def NormalizePOS(self, sRawPOS):
        """Normalize the POS and set it.  We require that the sRawPOS be
           one of a pre-defined set of POS names, which are given as keys in
           the dictionary dictPOSNames (in the included file <CodesFile>).
           If it is one of the keys, then we return the POS, which comes from using
           the canonical name as a key in dictPOSs.  (More than one raw POS may correspond
           to a single canonical name.)
           For non-legit POSs (or missing POSs), we return None."""
        if sRawPOS == '':
            return None  #This case probably doesn't happen
        if string.find(sRawPOS, ' ') != -1:
            #If we get here, sRawPOS contains a space char.  We could replace that with an underscore,
            # but the space probably indicates a comment, and we don't know what to do with that.
            # So we reject the record:
            raise Lexeme.POSSpaceCharWarning, self.sErrorMsgString
            return None
        else:
            try:
                sCanonicalPOS = dictPOSNames[sRawPOS]
            except KeyError:
                raise Lexeme.UnknownPOSWarning, (sRawPOS, self.sErrorMsgString)
                return None
            #At this point, we have a canonical POS name.  We want to return either the inflected
            # POS or a dummy POS for uninflected words:
            if sCanonicalPOS in sInflectedCategories:
                return dictPOSs[sCanonicalPOS]
            else:
                return UninflectedPOS


    def SetUnderlyingForm(self, sRawAllomorphClass):
        """For a few verb allomorph classes, the underlying form is different from the
           form in the lexicon.  The stem is adjusted here to be the underlying form,
           or a slightly different form.  (But citation and underlying form will be
           the same length.)"""
        if sRawAllomorphClass == "class-4a(oni)" or sRawAllomorphClass == "class-4a(pass)":
            #Verbs of paradigm class 4a(oni) have an underlying long stem-final vowel,
            # and at the same time they cannot be distinguished from the other
            # 4a verbs phonologically.  (All the 4a(oni) verbs end in /i/,
            # but so do some 4a verbs that are not 4a(oni).)  This length
            # will be removed for the lexical stem in VerbStems.xfst.
            #Verbs of class 4a(pass) also have an underlying long stem-final vowel.
            # While verbs of this subclass of 4a can be distinguished underlyingly
            # (they all end in /lo/), for consistency we treat the underlying length
            # here too, like class 4a(oni).
            self.sUnderlyingForm = self.sCitationForm + "%:"
            #...but the two have to be the same length, lest flag diacritics get offset;
            # so we add a '0' to the upper side:
            self.sCitationForm   = self.sCitationForm + "0"
        else:
            #Default case:
            self.sUnderlyingForm = self.sCitationForm
        return


    def ModPOSwithSubclass(self, sRawSubclass):
        """This procedure can modify self's POS to include the subclass.
           The subclass is relevant only for nouns.
           The POS must be normalized before the paradigm class.
           Returns the POS (possibly modified to include the subclass)."""
        #Figure out the subclass, if any:
        if (sRawSubclass == '') or (sRawSubclass == None):
            #Empty subclass, POS is unchanged
            return self.POS
        else:
            #Non-empty subclass, maybe add it to existing POS.
            # First strip out any space chars (because the use of space chars tend to be inconsistent):
            sRawSubclass = sRawSubclass.replace(" ", "")
            #If this POS has subclasses, look up the xfst subclass code:
            SubClasses = self.POS.Subclasses()
            if SubClasses != None:
                try:
                    #Subclass is a subcategory of the POS, so append it to the existing POS:
                    sCompositePOS = self.POS.sName + SubClasses[sRawSubclass]
                       #If this is a noun, self.POS will start out 'N', which we could
                       # hard code on the right-hand side of the above '='.
                       # But in the future this may change, so we use self.POS.
                    #Then change our POS membership to this composite POS:
                    return self.NormalizePOS(sCompositePOS)  #Treat the paradigm class as part of the noun POS
                except KeyError:
                    raise Lexeme.BadSubclassWarning, (sRawSubclass, self.sErrorMsgString)
                    #Return the unaltered POS, which will at least allow us to parse some forms:
                    return self.POS
            else:
                #Not a noun, and we don't know what to do with it; so ignore it, but
                # raise a warning:
                raise Lexeme.BadSubclassWarning, (sRawSubclass, self.sErrorMsgString)
                #But return the unaltered POS, which will at least allow us to parse some forms:
                return self.POS

    def InterpretIrregularities(self, sIrregularities):
        """Unclear what this will do in the long run.  For the time being, the only thing it knows
           to do is to add a flag diacritic to nouns that have no plural form, so that they can't
           take a plural suffix.
           No return value needed."""
        if sIrregularities == 'Singular' \
        or sIrregularities == 'Singular(mass)' \
        or sIrregularities == 'Singular(dom)' \
        or sIrregularities == 'Undetermined':
            #This is a noun that only appears in the singular, so add a flag diacritic:
            self.FlagDiacritics = self.FlagDiacritics + "@U.NUMBER.SG@"
               #Note no spaces!
        elif sIrregularities == 'Plural entry':
            sys.stderr.write("second clause, sIrregularities ='%s'\n" %sIrregularities)
            #Irregular plural form:
            self.FlagDiacritics = self.FlagDiacritics + "@U.NUMBER.PL@"


    def NormalizeStemClass(self, sRawStemClass):
        """Normalize stem class.
           For verbs only.  The lexicon uses a large set (in the \infv field), which we map to
           a smaller pre-defined xfst-particular set.  The codes are collected into the sRawStemClass,
           tokenized (if there's > 1), and normalized.
           We return the tags in xfst format, i.e. with a leading '^' and no space chars.
           (If there is > 1, we concatenate them; each will have the leading '^'.)
           If there is no stem class, we return an empty string.
           The POS and shape must be normalized before the stem class."""
        #Figure out the subclass, if any:
        if (sRawStemClass == '') or (sRawStemClass == None):
            #Empty class
            return ''    #Rather than None
        else:
            #Non-empty subclass, normalize it.
            #Look up the xfst subclass code:

            dictPOSStemClasses = self.POS.StemClasses()
            if dictPOSStemClasses == None:
                #This POS doesn't list any stem classes, so we shouldn't have found one
                # in the lex entry.  So raise a warning:
                raise Lexeme.UnexpectedStemClassWarning, (sRawStemClass, self.POS.sName, self.sErrorMsgString)
                #...but return an empty stem class anyway:
                return ''
            else:
                try:
                    return dictPOSStemClasses[sRawStemClass]
                except KeyError:
                    raise Lexeme.BadStemClassWarning, (sRawStemClass, self.sErrorMsgString)
                    return None


    def NormalizeAllomorphClass(self, sRawAllomorphClasses):
        """Normalize allomorph class(es).
           Similar to normalizing stem class, but uses a different set of standard codes.
           Also, a lexeme may have > 1 allomorph class.
           For verbs and nouns, the codes are carried by the \allomorph field.
           The codes are collected into the sRawAllomorphClasses, tokenized
           (if there's > 1) and normalized.
           We return the tags in xfst format, i.e. with a leading '^' and no space chars.
           (If there is > 1, we concatenate them; each will have the leading '^'.)
           If there is no allomorph class, we return an empty string.
           The POS and shape must be normalized before the allomorph class."""
        #Figure out the subclass, if any:
        if (sRawAllomorphClasses == '') or (sRawAllomorphClasses == None):
            #Empty class
            return ''    #Rather than None
        else:
            #Non-empty subclass, normalize it.
            #Look up the xfst subclass code.
            # First get the list of possible classes for this lexeme's POS:
            dictPOSAllomorphClasses = self.POS.AllomorphClasses()
            if dictPOSAllomorphClasses == None:
                #This POS doesn't list any allomorph classes, so we shouldn't have found one
                # in the lex entry.  So raise a warning:
                raise Lexeme.UnexpectedAllomorphClassWarning, (sRawAllomorphClasses, self.POS.sName, self.sErrorMsgString)
                #...but return an empty allomorph class anyway:
                return ''
            else:
                #Unlike NormalizeParadigmClass(), can't do dictPOSAllomorphClasses[sRawAllomorphClasses]
                # here, because we may have > 1 code in sRawAllomorphClasses.  So we instead have to
                # step through sRawAllomorphClasses looking for the codes, until we run out of string:
                sRawAllomorphClasses = sRawAllomorphClasses.lstrip() # = classes in lex entry
                sXFSTAllomorphClasses = ''                           # = classes to be used in xfst
                sAllomorphKeys = dictPOSAllomorphClasses.keys()      # = classes known for this POS
                while sRawAllomorphClasses != '':
                    #Go through the keys looking for one that is the left-hand end of this lexeme's allomorph codes:
                    # (We can't just tokenize the lexeme's keys, because some allomorph codes contain a space char.)
                    sXFSTCode = ''
                    for sKey in sAllomorphKeys:
                        if sRawAllomorphClasses.startswith(sKey):
                            #We found a raw code, use it to find the xfst code:
                            sXFSTCode = dictPOSAllomorphClasses[sKey]
                            #...and remove this raw code from the codes to be found for this lex entry,
                            # plus whitespace that separated this code from the next one (if any):
                            sRawAllomorphClasses = sRawAllomorphClasses[len(sKey):].lstrip()
                            break #drop out of the loop
                    #Now make sure we found a code (if we didn't, without this check we'd infinitely loop):
                    if sXFSTCode == '':
                        #If we get here, we didn't find a code; signal an error:
                        raise Lexeme.BadAllomorphClassWarning, (sRawAllomorphClasses, self.sErrorMsgString)
                        return None
                    else:   #The else is superfluous, but...
                        #Add the new code to the existing codes:
                        sXFSTAllomorphClasses = sXFSTAllomorphClasses + sXFSTCode
                #When we get here, we've used up all of sRawAllomorphClasses, and we're ready to return:
                return sXFSTAllomorphClasses



    def NormalizeParadigmClass(self, sRawParadigmClass):
        """Normalize paradigm class.  This is only relevant to nouns (i.e. this is
           a declension class).  Return the paradigm class, if any.  If there is
           no paradigm class, we return an empty string."""
        if sRawParadigmClass == None:
            return ''
        else:
            #There was a field for a paradigm class
            if sRawParadigmClass == '':
                #...but it was empty
                return ''  #Rather than None
            #If we get here, we have a non-empty paradigm class.
            elif self.IsA("N"):  #For nouns, the lexicon has two declension classes: native, and Spanish loans.
                                 # The former are unmarked, the latter are marked by having a \loan field.
                                 # The contents of that field (= sRawParadigmClass) are irrelevant--it is
                                 # enough that they have the field.  So we just attach '"@P.STRATUM.SPANISH@".
                                 # (This means that we mark Spanish loanwords, but do not mark non-Spanish
                                 # loans; this in turn requires the D(isallow) and R(equire) diacritics on affixes.)
                    return "@P.STRATUM.SPANISH@"
            else:
                #We shouldn't get here--it means we encountered a non-null paradigm class on a lexeme
                # that is not a noun.  Issue a warning:
                try:
                        raise Lexeme.BadParadigmClassWarning, (sRawParadigmClass, self.sErrorMsgString)
                #But continue, on the assumption that it's better to ignore such a code than throw
                # away the entire lexeme:
                except Lexeme.BadParadigmClassWarning:
                        pass
                return ''


    #-----------------Output functions----------------------
    def Write_lexc_format(self, OutputStream):
        """Write self to OutputStream in lexc format."""
        #Ensure citation and underlying form are the same number of chars, not counting the escape char '%'
        # (otherwise the flag diacritics can get offset, which will cause xfst to complain).  (An assumption
        # here is that the char '%' is not itself present in one of the strings, represented as '%%'.)
        if len(self.sCitationForm.replace('%', '')) != len(self.sUnderlyingForm.replace('%', '')):
            raise Lexeme.DifferentLengthsWarning, (self.sCitationForm, self.sUnderlyingForm)
        else:
            #Upper side is citation form, lower side is underlying form + stem class + allomorph code + paradigm class;
            # any flag diacritics are written on both sides.  (Flags are written adjacent to the stem, *before* any stem class
            # or allomorph class, so that the presence of the latter on the lower side doesn't cause the flags to be offset
            # between the upper and lower side.  Since the paradigm class is also represented by a flag diacritic, it is also
            # written before the stem and allomorph class, for the same reason.)
            OutputStream.write('\t%s%s%s:%s%s%s%s%s\t#;\n'
                               %(self.sCitationForm,   self.FlagDiacritics, self.sParadigmClass,
                                 self.sUnderlyingForm, self.FlagDiacritics, self.sParadigmClass, self.sStemClass, self.sAllomorphClasses)
                              )

    #-----------------Is-A functions------------------------
    def IsA(self, sGenericPOS):
        """Return true if self's POS is a subcategory of sGenericPOS
           (e.g. V1 is a subclass of V)."""
        return self.POS.IsA(sGenericPOS)


    def IsInflected(self):
        "Return True if the word is inflected (based on its POs), False otherwise. \
         Signal an error if the POS is not set."
        if self.POS == None:
            #Error--it's not set yet
            raise Lexeme.POSNotSetWarning, self.sErrorMsgString
        else:
            #We have a POS, so proceed:
#            sys.stderr.write(" %s %s\n" %(self.POS, self.POS not in sInflectedCategories))
            return self.POS.IsInflected


    #---------------Lexeme class warnings-------------
    #All of these are treated as Warnings, in the sense that we continue processing the rest of
    # the input file.  That means that there has to be an 'except' clause for them somewhere
    # (currently, most are handled by a single 'except' clause in the main loop).

    class MissingStemForm(Exception):
        #The msg says the citation form is not known for the current dialect, but in fact we only
        # handle Oapan at present.
        def __init__(self, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with no citation form for current dialect: %s" %sErrorMsgString)

    class EmptyStemWarning(Exception):
        def __init__(self, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with empty stem: %s" %sErrorMsgString)

    class PhrasalLexemeWarning(Exception):
        #This is not an error condition, but we issue the warning just to keep track of how many
        # lex entries we're skipping because they're phrasal
        def __init__(self, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with phrasal lexeme: %s" %sErrorMsgString)

    class StemCharsWarning(Exception):
        def __init__(self, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with invalid chars in stem: %s" %sErrorMsgString)

    class MissingPOSWarning(Exception):
        def __init__(self, sErrorMsgString):
            ErrOut.write("Warning: Skipping record which lacks POS: %s" %sErrorMsgString)

    class POSSpaceCharWarning(Exception):
        def __init__(self, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with space char in POS: %s" %sErrorMsgString)

    class UnknownPOSWarning(Exception):
        def __init__(self, sRawPOS, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with unknown POS '%s': %s" %(sRawPOS, sErrorMsgString))

    class BadSubclassWarning(Exception):
        def __init__(self, sRawSubclass, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with invalid subclass '%s': %s" %(sRawSubclass, sErrorMsgString))

    class UnexpectedStemClassWarning(Exception):
        def __init__(self, sRawStemClass, sPOS, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with invalid stem class '%s' for POS %s: %s" %(sRawStemClass, sPOS, sErrorMsgString))

    class BadStemClassWarning(Exception):
        def __init__(self, sRawStemClass, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with unknown stem class '%s': %s" %(sRawStemClass, sErrorMsgString))

    class UnexpectedAllomorphClassWarning(Exception):
        def __init__(self, sRawAllomorphClass, sPOS, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with invalid allomorph class '%s' for POS %s: %s" %(sRawAllomorphClass, sPOS, sErrorMsgString))

    class BadAllomorphClassWarning(Exception):
        def __init__(self, sRawAllomorphClass, sErrorMsgString):
            ErrOut.write("Warning: Skipping record with unknown allomorph class '%s': %s" %(sRawAllomorphClass, sErrorMsgString))

    class BadParadigmClassWarning(Exception):
        def __init__(self, sRawParadigmClass, sErrorMsgString):
            ErrOut.write("Warning: Ignoring unknown paradigm class '%s': %s" %(sRawParadigmClass, sErrorMsgString))

    class DifferentLengthsWarning(Exception):
        def __init__(self, sCitationForm, sUnderlyingForm):
            ErrOut.write("Warning: Citation form '%s' and underlying form '%s' are not of equal lengths.\n\n" %(sCitationForm, sUnderlyingForm))

    class POSNotSetWarning(Exception):
        #This shouldn't happen, since we control when the POS is set (it has to be set before
        # the subclass can be set).
        def __init__(self, sErrorMsgString):
            ErrOut.write("Warning: POS not set for this record: %s" %sErrorMsgString)


#--------------------------Class POS---------------------------
class POS:
    """Keeps track of a (normalized) POS and its lexemes.
    """

    def __init__(self, sName):
        #Done at program initialization.
        self.sName = sName
        self.Lexemes = []
        self.IsInflected = self.sName in sInflectedCategories


    def IsA(self, sGenericPOS):
        """Return true iff self is a subcategory of sGenericPOS
           (e.g. V1 is a subclass of V)."""
        if sPOSSuperclass.has_key(self.sName):
            #sGenericPOS has a superclass POS, see if it's the right one:
            return sPOSSuperclass[self.sName] == sGenericPOS
        else:
            #sGenericPOS doesn't have a superclass POS, so return False:
            return False


    def WriteXFSTLoaderCode(self, XFSTStream):
        "If this POS has any lexemes, then write the load information for this POS to the xfst file."
        if self.Lexemes != []:
            XFSTStream.write('read lexc %s.lexc\ndefine %s\n\n' %(self.sName, self.sName))


    def SuperPOS(self):
        "Return in string form the superclass of this POS, if any; else None."
        if sPOSSuperclass.has_key(self.sName):
            return sPOSSuperclass[self.sName]
        else:
            return None


    def Subclasses(self):
        "Return a list of subclasses of this POS, if any."
        if dictSubclasses.has_key(self.sName):
            return dictSubclasses[self.sName]
        else:
            return None


    def StemClasses(self):
        "Return the dictionary of stem classes for this POS, if any."
        #Get the POS's superclass name, if any:
        sStemKey = self.SuperPOS()
        if sStemKey == None:
            sStemKey = self.sName
        #...and see if this POS (or its superclass) has any stem classes:
        if dictStemClasses.has_key(sStemKey):
            return dictStemClasses[sStemKey]
        else:
            return None

    def AllomorphClasses(self):
        "Return the dictionary of allomorph classes for this POS, if any."
        #Get the POS's superclass name, if any:
        sAllomorphKey = self.SuperPOS()
        if sAllomorphKey == None:
            sAllomorphKey = self.sName
        #...and see if this POS (or its superclass) has any allomorph classes:
        if dictAllomorphClasses.has_key(sAllomorphKey):
            return dictAllomorphClasses[sAllomorphKey]
        else:
            return None


    def WriteLexcFile(self):
        """If this POS has any lexemes, then write out a lexc file with those lexemes."""
        if __debug__:
            sys.stderr.write("POS %s, # lexemes = %i\n" %(self.sName, len(self.Lexemes)))
        if self.Lexemes != []:
            (UTF8Encode, UTF8Decode, UTF8Reader, UTF8Writer) = codecs.lookup('utf-8')

            POSLexiconOut = UTF8Writer(open(self.sName + '.lexc', 'w'))
            #Header for this lexc file:
            POSLexiconOut.write('!Generated from %s on %s by NahTab2MultLexc.py.\n\n'
                                %(LexiconIn.name, time.strftime('%d %B %Y at time %H:%M', time.localtime()))
                               )
            #Write out any stem classes and allomorph classes, plus diacritic flags, as Multichar symbols:
            dictPOSStemClasses = self.StemClasses()
            dictPOSAllomorphClasses = self.AllomorphClasses()
            if dictPOSStemClasses != None or dictPOSAllomorphClasses != None or sDiacriticFlags != None:
                #If we get here, there are stem classes or allomorph classes for this POS, or diacritic flags
                # (we don't keep track of the latter by POS), so write them out:
                POSLexiconOut.write('Multichar_Symbols\n')
                if sDiacriticFlags != None:
                    for sDiacriticFlag in sDiacriticFlags:
                        POSLexiconOut.write('\t%s\n' %sDiacriticFlag)
                if dictPOSStemClasses != None:
                    for sStemClass in dictPOSStemClasses.values():
                        POSLexiconOut.write('\t%s\n' %sStemClass)
                if dictPOSAllomorphClasses != None:
                    for sAllomorphClasses in dictPOSAllomorphClasses.values():
                        POSLexiconOut.write('\t%s\n' %sAllomorphClasses)
                if self.IsA("N"):  #Noun
                    #Also write out "@P.STRATUM.SPANISH@" (hard coded for now, used for loan words;
                    # only relevant for nouns at this point, since the declension of Spanish loan nouns
                    # is different):
                    POSLexiconOut.write('\t@P.STRATUM.SPANISH@\n')
                POSLexiconOut.write('\n')
            #...write out lexicon headers:
            POSLexiconOut.write('LEXICON Root\n\t' + self.sName + ' ;\n\nLEXICON ' + self.sName + '\n')
            #...and write out all the lexemes for this POS:
            for Lexeme in self.Lexemes:
                #Upper side is citation form + paradigm class,
                # lower side is underlying form + allomorph class + paradigm class:
                try:
                    Lexeme.Write_lexc_format(POSLexiconOut)
                except Lexeme.DifferentLengthsWarning:
                    pass #The error msg is written out to the error stream, and we just skip this lexeme
            POSLexiconOut.close()



#-------------------------------------Main-----------------------------
#--------------Initialization:
(CodesFile, LexiconIn, XfstOut, ErrOut) = OpenInputOutput()
#Get the constants from the file specified on the command line:
try:
    execfile(CodesFile)
except IOError:
    #Any other error will get its ordinary error msg and cause an exit.
    sys.stderr.write("Error: Unable to read  language-specific initialization file %s.\n"
                     %CodesFile)
    sys.exit(1)
#Create a dictionary struc to hold a set of POS instances, one for each inflected POS:
dictPOSs = {}
    #Finding the POS instance for a word in the dictionary will be a two-step process:
    #  (1) Map the word's raw POS to a canonical POS name (more than one raw POS may map into
    #      the same canonical name)
    #  (2) Map the canonical name to a POS instance (one-to-one relation)
#Init this dict by creating a POS instance for each unique canonical POS name to the that dictionary
# which is an inflected POS:
for sPOSName in set(dictPOSNames.values()):
    if sPOSName in sInflectedCategories:
        #Create a POS instance for this inflected POS:
        dictPOSs[sPOSName] = POS(sPOSName)
#Also create a dummy POS for uninflected forms, which will go into a separate file:
UninflectedPOS = POS('Uninflected')

#Create a list of diacritic flags we encounter, so these can be declared in the lexc files
# as Multichar_Symbols.  For now, this list is hard-coded:
sDiacriticFlags = ["@U.NUMBER.SG@"]

#Comment header in xfst file:
XfstOut.write('!Generated from %s by NahTab2MultLexc.py on %s.\n'
              %(LexiconIn.name, time.strftime('%d %B %Y at time %H:%M', time.localtime()))
             )

#--------------Read in the lexicon file:
domLexicon = xml.dom.minidom.parse(LexiconIn)
LexiconIn.close()

#--------------Process records:
#sAllTags is used to create an error msg string:
sAllTags = sCitationFormTags + sPOSTags + sSubclassTags + sAllomorphTags

#Convert from XML DOM form to Lexeme objects, which will store themselves
# on the lexeme's POS for inflected POSs, or else on the dummy UninflectedPOS.
#Count records; if we're debugging, we'll output a msg every 10 records:
iRecordNum = 0
for domLexemeNode in domLexicon.getElementsByTagName(sLexemeTag):
    try:
        Lexeme(domLexemeNode)
    except (Lexeme.MissingStemForm, Lexeme.EmptyStemWarning, Lexeme.PhrasalLexemeWarning, Lexeme.StemCharsWarning,
            Lexeme.MissingPOSWarning, Lexeme.POSSpaceCharWarning, Lexeme.UnknownPOSWarning, Lexeme.POSNotSetWarning,
            Lexeme.BadSubclassWarning,
            Lexeme.UnexpectedStemClassWarning, Lexeme.BadStemClassWarning,
            Lexeme.UnexpectedAllomorphClassWarning, Lexeme.BadAllomorphClassWarning,
            Lexeme.BadParadigmClassWarning):
            #Lexeme.DifferentLengthsWarning is not triggered until we attempt output (see WriteLexcFile)
        pass  #Any error msg has already been output, so just go on to the next record
    iRecordNum = iRecordNum + 1
    if __debug__ and iRecordNum % 10 == 0:
            #Update the progress msg every ten entries:
            sys.stderr.write("\rNahTab2MultLexc.py: Processing record %i" %iRecordNum)
if __debug__:
    sys.stderr.write("\rNahTab2MultLexc.py: Processed %i records.\n" %iRecordNum)
#Done with the dom, release its memory (see documentation for xml.dom.minidom for
# why this is necessary--gc may not get it):
domLexicon.unlink()


#Now output the records in lexc files, one file for each POS that takes inflection
# (but only if there actually are lexemes for a given POS; that condition is checked
# by the POS object before writing the file).
if __debug__:
    sys.stderr.write("\nNahTab2MultLexc.py: Creating %i lexc files for POSs (including uninflected).\n"
                     %(len(sInflectedCategories) + 1)
                    )
InflectedPOSs = dictPOSs.values()
for POS in InflectedPOSs + [UninflectedPOS]:
    if __debug__:
        sys.stderr.write("NahTab2MultLexc.py: Creating lexc file for POS %s\n" %POS.sName)
    POS.WriteXFSTLoaderCode(XfstOut)
    POS.WriteLexcFile()

XfstOut.close()
ErrOut.close()

