#!/usr/bin/env python
## -*- coding: utf-8 -*-
#
#Combine two files.  The first file is an XML file that contains surface forms that
# have been generated (for now, as cells of a paradigm; this may find other uses).
# The second file is a tab-delimited file containing three columns:
#   <GlossParse>\t<UnderlyingForm>\t<SurfaceForm>
# The second file contains unique surface forms, whereas the first file may contain
# the same surface form > once.
#
# The output is an XML file identical to the first one, except that the surface forms
# have been replaced by a new XML tag containing the surface form, UR, and Gloss:
#    <parse sr="<SurfaceForm>" ur="<UnderlyingRepresentation>" gl="<Gloss>"\>
# Note: If the surface form is ambiguous, we will get multiple parses.  These are given
# in separate <parse\> tags.  The alternative would be to generate the underlying and
# gloss forms in the same way we generate the surface form.  However, where the paradigm
# cell generates more than one form, there is no guarantee that the three forms would be
# generated in the same order.  It will often be the case that a paradigm cell contains
# more than one form (e.g. multiple future tense forms), but less often will there be
# ambiguity of parsing--at least in Nahuatl.  The ambiguity could be reduced by compiling
# separate transducers for each root that we want to test.
#
# Arguments: -x <InputXMInXMLFile>
#            -p <Input3ColumnFile>
#            -u (=> UTF-8)
# The first two arguments are obligatory, the third is optional (and if
# not given, the input is assumed to be non-UTF8, presumably ISO).
##FIX: third arg may be superfluous, since the XML file must specify its encoding.

# Input: The two files, as described above.
# Output: An XML file output to stdout, as described.
# Exit code 1 if there is an error (a surface form in the XML file that couldn't
#     be parsed--this should not be possible; or failure to open one of the input
#    files); otherwise 0.

import sys
import getopt
import codecs   #For UTF-8


def OpenInput():
    """Open the two input files, and the output and error streams."""

    #Default is non-Unicode:
    bUTF8 = False

    #Look for command line args:
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'ux:p:')
    except getopt.GetoptError:
        sys.stderr.write( 'Usage: SR2Parses.py -x <Input XML file> -p <Input 3-column file> (-u)\n')
        sys.exit(1)

    #Init file names for later error checking:
    InXMLFile = ""
    In3ColumnFile = ""

    for o in opts:
        if o[0] == '-x':
            InXMLFile = o[1]
        elif o[0] == '-p':
            In3ColumnFile = o[1]
        elif o[0] == '-u':
            bUTF8 = True
            (UTF8Encode, UTF8Decode, UTF8Reader, UTF8Writer) = codecs.lookup('utf-8')
               # We don't use the UTF8Encode or UTF8Decode, but they are returned by codecs.lookup()

    #Make sure both input files exist:
    if InXMLFile == "":
        sys.stderr.write( 'Usage: Must specify name of XML input file!\n')
        sys.exit(1)
    if In3ColumnFile == "":
        sys.stderr.write( 'Usage: Must specify name of 3-column input file!\n')
        sys.exit(1)
    #If we get here, we were given the names of the files; open them:
    try:
        if bUTF8:
            InXMLStream = UTF8Reader(open(InXMLFile, 'r'))
        else:
            InXMLStream = open(InXMLFile, 'r')
    except IOError:
        sys.stderr.write("Error: Cannot open file %s\n" %InXMLFile)
        sys.exit(1)

    try:
        if bUTF8:
            In3ColumnStream = UTF8Reader(open(In3ColumnFile, 'r'))
        else:
            In3ColumnStream = open(In3ColumnFile, 'r')
    except IOError:
        sys.stderr.write("Error: Cannot open file %s\n" %In3ColumnFile)
        sys.exit(1)

    #Also open the output stream and error streams.  Output is to stdout,
    # but we still have to assign it to a stream here, in case we're writing Unicode:
    if bUTF8:
        OutStream = UTF8Writer(sys.stdout)
        ErrStream = UTF8Writer(sys.stderr)
    else:
        outStream = sys.stdout
        ErrStream = sys.stderr

    return (InXMLStream, In3ColumnStream, OutStream, ErrStream)


def ParseLine(sLine):
    """A trivial function to parse sLine into three parts at tab chars,
       removing any other initial or final whitespace.
       Return the three parts, or -1 if the line could not be split."""
    try:
        (sPart1, sPart2, sPart3) = map(lambda(s): s.strip(), sLine.split("\t", 2))
        return (sPart1, sPart2, sPart3)
    except:
        return -1


#--------------MAIN--------------
iErrorStatus = 0  #Hope for the best

#--------------Process records:
#Init the pipeline:
(InXMLStream, In3ColumnStream, OutStream, ErrStream) = OpenInput()

#First read all the 3-column parses into a dictionary struc, and index by
# the Gl.
dictParses = {}
for sRecord in In3ColumnStream.readlines():
    (sGl, sUR, sSR) = ParseLine(sRecord)
    if dictParses.has_key(sGl):
       #Gl is ambiguous, add the new parse:
       dictParses[sGl].append((sUR, sSR))
    else:
       dictParses[sGl] =     [(sUR, sSR)]

#...and then read the XML file in, and transduce it:
sXMLLine = InXMLStream.readline().rstrip()
   #use rstrip() instead of strip() to preserve indent on left-hand side
while sXMLLine <> "":
    if sXMLLine.find("<") <> -1:
       #An XML tag line, just echo it to output:
       OutStream.write("%s\n" %sXMLLine)
    else:
       #It's a gloss line:
       sGl = sXMLLine.strip()
       if dictParses.has_key(sGl):
          #This should always be the case.  We'll produce a line with all three fields for
          # every parse:
          for (sUR, sSR) in dictParses[sGl]:
            OutStream.write('          <parse sr="%s" ur="%s" gl="%s"/>\n' %(sSR, sUR, sGl))
       else:
          #This should not happen--it means one of the words that was generated
          # couldn't be parsed.
          ErrStream.write("Error: Unable to generate from %s\n" %sGl)
          iErrorStatus = 1
    sXMLLine = InXMLStream.readline().rstrip()

sys.exit(iErrorStatus)
